@@ -27,6 +27,8 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <errno.h>
+#include <sys/stat.h>
 
 #include <glib.h>
 
@@ -40,6 +42,8 @@
 
 #include "atmodem.h"
 
+#define TUN_SYSFS_DIR "/sys/devices/virtual/misc/tun"
+
 #define STATIC_IP_NETMASK "255.255.255.255"
 
 static const char *none_prefix[] = { NULL };
@@ -65,7 +69,13 @@ struct gprs_context_data {
 	void *cb_data;                                  /* Callback data */
 };
 
-static void ppp_connect(const char *interface, const char *ip,
+static void ppp_debug(const char *str, void *data)
+{
+	ofono_info("%s: %s", (const char *) data, str);
+}
+
+static void ppp_connect(const char *interface, const char *local,
+			const char *remote,
 			const char *dns1, const char *dns2,
 			gpointer user_data)
 {
@@ -73,12 +83,17 @@ static void ppp_connect(const char *interface, const char *ip,
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 	const char *dns[3];
 
+	DBG("");
+
 	dns[0] = dns1;
 	dns[1] = dns2;
 	dns[2] = 0;
 
+	ofono_info("IP: %s", local);
+	ofono_info("DNS: %s, %s", dns1, dns2);
+
 	gcd->state = STATE_ACTIVE;
-	CALLBACK_WITH_SUCCESS(gcd->up_cb, interface, TRUE, ip,
+	CALLBACK_WITH_SUCCESS(gcd->up_cb, interface, TRUE, local,
 					STATIC_IP_NETMASK, NULL,
 					dns, gcd->cb_data);
 }
@@ -92,7 +107,6 @@ static void ppp_disconnect(GAtPPPDisconnectReason reason, gpointer user_data)
 
 	g_at_ppp_unref(gcd->ppp);
 	gcd->ppp = NULL;
-	g_at_chat_resume(gcd->chat);
 
 	switch (gcd->state) {
 	case STATE_ENABLING:
@@ -109,6 +123,12 @@ static void ppp_disconnect(GAtPPPDisconnectReason reason, gpointer user_data)
 
 	gcd->active_context = 0;
 	gcd->state = STATE_IDLE;
+	/*
+	 * If the channel of gcd->chat is NULL, it might cause
+	 * gprs_context_remove get called and the gprs context will be
+	 * removed.
+	 */
+	g_at_chat_resume(gcd->chat);
 }
 
 static gboolean setup_ppp(struct ofono_gprs_context *gc)
@@ -116,6 +136,8 @@ static gboolean setup_ppp(struct ofono_gprs_context *gc)
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 	GAtIO *io;
 
+	DBG("");
+
 	io = g_at_chat_get_io(gcd->chat);
 
 	g_at_chat_suspend(gcd->chat);
@@ -128,6 +150,9 @@ static gboolean setup_ppp(struct ofono_gprs_context *gc)
 		return FALSE;
 	}
 
+	if (getenv("OFONO_PPP_DEBUG"))
+		g_at_ppp_set_debug(gcd->ppp, ppp_debug, "PPP");
+
 	g_at_ppp_set_credentials(gcd->ppp, gcd->username, gcd->password);
 
 	/* set connect and disconnect callbacks */
@@ -145,6 +170,8 @@ static void at_cgdata_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct ofono_gprs_context *gc = user_data;
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 
+	DBG("ok %d", ok);
+
 	if (!ok) {
 		struct ofono_error error;
 
@@ -168,6 +195,8 @@ static void at_cgdcont_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 	char buf[64];
 
+	DBG("ok %d", ok);
+
 	if (!ok) {
 		struct ofono_error error;
 
@@ -200,6 +229,8 @@ static void at_gprs_activate_primary(struct ofono_gprs_context *gc,
 	char buf[OFONO_GPRS_MAX_APN_LENGTH + 128];
 	int len;
 
+	DBG("cid %u", ctx->cid);
+
 	gcd->active_context = ctx->cid;
 	gcd->up_cb = cb;
 	gcd->cb_data = data;
@@ -222,12 +253,12 @@ static void at_gprs_activate_primary(struct ofono_gprs_context *gc,
 }
 
 static void at_gprs_deactivate_primary(struct ofono_gprs_context *gc,
-					unsigned int id,
+					unsigned int cid,
 					ofono_gprs_context_cb_t cb, void *data)
 {
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 
-	DBG("");
+	DBG("cid %u", cid);
 
 	gcd->state = STATE_DISABLING;
 	gcd->down_cb = cb;
@@ -241,9 +272,20 @@ static int at_gprs_context_probe(struct ofono_gprs_context *gc,
 {
 	GAtChat *chat = data;
 	struct gprs_context_data *gcd;
+	struct stat st;
 
-	gcd = g_new0(struct gprs_context_data, 1);
-	gcd->chat = chat;
+	DBG("");
+
+	if (stat(TUN_SYSFS_DIR, &st) < 0) {
+		ofono_error("Missing support for TUN/TAP devices");
+		return -ENODEV;
+	}
+
+	gcd = g_try_new0(struct gprs_context_data, 1);
+	if (!gcd)
+		return -ENOMEM;
+
+	gcd->chat = g_at_chat_clone(chat);
 
 	ofono_gprs_context_set_data(gc, gcd);
 
@@ -256,12 +298,14 @@ static void at_gprs_context_remove(struct ofono_gprs_context *gc)
 
 	DBG("");
 
-	if (gcd->state != STATE_IDLE) {
+	if (gcd->state != STATE_IDLE && gcd->ppp) {
 		g_at_ppp_unref(gcd->ppp);
 		g_at_chat_resume(gcd->chat);
 	}
 
 	ofono_gprs_context_set_data(gc, NULL);
+
+	g_at_chat_unref(gcd->chat);
 	g_free(gcd);
 }
 
